#include <stdio.h>
/* rand, abs */
#include <stdlib.h>
/* time */
#include <time.h>
/* strcmp */
#include <string.h>

#include "array.h"
#include "digital-fov-other.h"
#include "digital-fov.h"
#include "util-example.h"
#include "util.h"

static int calculate_fov(int **map, int map_size_x, int map_size_y,
                         int center_x, int center_y, int radius,
                         int (*fov_func)(int **, int, int,
                                         int **,
                                         int, int, int),
                         int whole_map, int circle_fov);

static int
calculate_fov(int **map, int map_size_x, int map_size_y,
              int center_x, int center_y, int radius,
              int (*fov_func)(int **, int, int,
                              int **,
                              int, int, int),
              int whole_map, int circle_fov)
{
  int x;
  int y;
  int c;
  int x_start;
  int x_end;
  int y_start;
  int y_end;
  int **map_fov = NULL;

  map_fov = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map_fov  == NULL)
  {
    return 1;
  }

  (*fov_func)(map,  map_size_x, map_size_y,
              map_fov,
              center_x, center_y, radius);
  if (circle_fov)
    trim_circle(map_fov, radius);

  if (whole_map)
  {
    x_start = 0;
    x_end = map_size_x - 1;
    y_start = 0;
    y_end = map_size_y - 1;
  }
  else
  {
    x_start = center_x - radius;
    if (x_start < 0)
      x_start = 0;
    x_end = center_x + radius;
    if (x_end > map_size_x - 1)
      x_end = map_size_x - 1;
    y_start = center_y - radius;
    if (y_start < 0)
      y_start = 0;
    y_end = center_y + radius;
    if (y_end > map_size_y - 1)
      y_end = map_size_y - 1;
  }

  if (whole_map)
    printf("; map size: %d x %d\n", map_size_x, map_size_y);
  printf("; center: (%d, %d)\n", center_x, center_y);

  for (y = y_start; y <= y_end; y++)
  {
    for (x = x_start; x <= x_end; x++)
    {
      if ((x == center_x) && (y == center_y))
      {
        c = '@';
      }
      else if ((distance(x, y, center_x, center_y) <= radius)
               && (map_fov[x - center_x + radius][y - center_y + radius]))
      {
        if (map[x][y] != 0)
          c = '#';
        else
          c = '.';
      }
      else
      {
        if (map[x][y] != 0)
          c = 'X';
        else
          c = 'x';
      }
      putchar(c);
    }

    putchar('\n');
  }

  array2_delete(map_fov, 2 * radius + 1, 2 * radius + 1);
  map_fov = NULL;

  return 0;
}

int
main(int argc, char *argv[])
{
  int i;
  int map_size_x_max;
  int map_size_y_max;
  int map_size_x;
  int map_size_y;
  int center_x;
  int center_y;
  int center_x_command_line;
  int center_y_command_line;
  int radius;
  int wall_rarity;
  int whole_map;
  int circle_fov;
  int **map = NULL;
  char *file_name = NULL;
  int (*fov_func)(int **, int, int,
                  int **,
                  int, int, int) = NULL;

  map_size_x_max = 1001;
  map_size_y_max = 1001;
  radius = 10;
  map_size_x = 2 * radius + 1;
  map_size_y = 2 * radius + 1;
  center_x = radius;
  center_y = radius;
  center_x_command_line = -1;
  center_y_command_line = -1;
  wall_rarity = 6;
  whole_map = 0;
  circle_fov = 0;
  fov_func = &digital_fov;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--map-file"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      file_name = argv[i + 1];
      i++;
    }
    else if (!strcmp(argv[i], "--fov"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      else if (!strcmp(argv[i + 1], "recursive"))
      {
        fov_func = &digital_fov;
        i++;
      }
      else if (!strcmp(argv[i + 1], "brute"))
      {
        fov_func = &digital_fov_brute;
        i++;
      }
      else if (!strcmp(argv[i + 1], "beam"))
      {
        fov_func = &digital_fov_beam;
        i++;
      }
      else if (!strcmp(argv[i + 1], "los"))
      {
        fov_func = &digital_fov_los;
        i++;
      }
      else if (!strcmp(argv[i + 1], "dummy"))
      {
        fov_func = &dummy_fov;
        i++;
      }
      else
      {
        fprintf(stderr, "main: unknown fov: %s\n", argv[i + 1]);
        return 1;
      }
    }
    else if (!strcmp(argv[i], "--radius"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&radius, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (radius < 0)
      {
        fprintf(stderr, "main: radius (%d) must be non-negative\n", radius);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--wall-rarity"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&wall_rarity, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int failed\n");
        return 1;
      }
      if (wall_rarity <= 0)
      {
        fprintf(stderr, "main: wall_rarity (%d) must be positive\n",
                wall_rarity);
        return 1;
      }
      i++;
    }
    else if (!strcmp(argv[i], "--center"))
    {
      if (i + 2 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      if (str_to_int(&center_x_command_line, argv[i + 1], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int(%s) failed\n", argv[i + 1]);
        return 1;
      }
      if (str_to_int(&center_y_command_line, argv[i + 2], 0) != 0)
      {
        fprintf(stderr, "main: str_to_int(%s) failed\n", argv[i + 2]);
        return 1;
      }
      if (center_x_command_line < 0)
      {
        fprintf(stderr, "main: center_x (%d) must be positive\n",
                center_x);
        return 1;
      }
      if (center_y_command_line < 0)
      {
        fprintf(stderr, "main: center_y (%d) must be positive\n",
                center_x);
        return 1;
      }
      i += 2;
    }
    else if (!strcmp(argv[i], "--whole-map"))
    {
      whole_map = 1;
    }
    else if (!strcmp(argv[i], "--circle-fov"))
    {
      circle_fov = 1;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      return 1;
    }
  }

  map = array2_new(map_size_x_max, map_size_y_max);
  if (map == NULL)
  {
    return 1;
  }

  srand((unsigned int) time(NULL));

  if (file_name != NULL)
  {
    if (read_map_from_file(file_name,
                           map, map_size_x_max, map_size_y_max,
                           &map_size_x, &map_size_y,
                           &center_x, &center_y, NULL,
                           0) != 0)
    {
      fprintf(stderr, "main: read_map_from_file failed\n");
      return 1;
    }
  }
  else
  {
    random_map(map, map_size_x_max, map_size_y_max,
               &map_size_x, &map_size_y,
               &center_x, &center_y, NULL,
               radius, wall_rarity);
  }

  if (center_x_command_line >= 0)
    center_x = center_x_command_line;
  if (center_y_command_line >= 0)
    center_y = center_y_command_line;
  if (grid_is_illegal(center_x, center_y, map_size_x, map_size_y))
  {
    fprintf(stderr, "center (%d, %d) is out of map\n", center_x, center_y);
    fprintf(stderr, "(map size is %d x %d)\n", map_size_x, map_size_y);
    array2_delete(map, map_size_x_max, map_size_y_max);
    map = NULL;
    return 1;
  }

  calculate_fov(map, map_size_x, map_size_y,
                center_x, center_y, radius,
                fov_func,
                whole_map, circle_fov);

  array2_delete(map, map_size_x_max, map_size_y_max);
  map = NULL;

  return 0;
}
