#include <stdio.h>
/* rand */
#include <stdlib.h>
/* time */
#include <time.h>

#include "array.h"
#include "digital-fov-other.h"
#include "digital-fov.h"

static int one_random_map(void);

static int
one_random_map(void)
{
  int x;
  int y;
  int c;
  int center_x;
  int center_y;
  int radius;
  int **map = NULL;
  int **map_fov_brute = NULL;
  int **map_fov_beam = NULL;
  int **map_fov_recursive = NULL;

  radius = 10;

  map = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map == NULL)
  {
    return 1;
  }
  map_fov_brute = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map_fov_brute == NULL)
  {
    array2_delete(map, 2 * radius + 1, 2 * radius + 1);
    map = NULL;
    return 1;
  }
  map_fov_beam = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map_fov_beam == NULL)
  {
    array2_delete(map, 2 * radius + 1, 2 * radius + 1);
    map = NULL;
    array2_delete(map_fov_brute, 2 * radius + 1, 2 * radius + 1);
    map_fov_brute = NULL;
    return 1;
  }
  map_fov_recursive = array2_new(2 * radius + 1, 2 * radius + 1);
  if (map_fov_recursive == NULL)
  {
    array2_delete(map, 2 * radius + 1, 2 * radius + 1);
    map = NULL;
    array2_delete(map_fov_brute, 2 * radius + 1, 2 * radius + 1);
    map_fov_brute = NULL;
    array2_delete(map_fov_beam, 2 * radius + 1, 2 * radius + 1);
    map_fov_beam = NULL;
    return 1;
  }

  for (x = 0; x <= 2 * radius; x++)
  {
    for (y = 0; y <= 2 * radius; y++)
    {
      if (rand() % 6 == 0)
        map[x][y] = 1;
      else
        map[x][y] = 0;
    }
  }

  center_x = radius;
  center_y = radius;
  digital_fov_brute(map, 2 * radius + 1, 2 * radius + 1,
                   map_fov_brute, center_x, center_y, radius);
  digital_fov_beam(map, 2 * radius + 1, 2 * radius + 1,
                   map_fov_beam, center_x, center_y, radius);
  digital_fov(map, 2 * radius + 1, 2 * radius + 1,
              map_fov_recursive, center_x, center_y, radius);

  for (y = center_y - radius; y <= center_y + radius; y++)
  {
    for (x = center_x - radius; x <= center_x + radius; x++)
    {
      if ((x == center_x) && (y == center_y))
      {
        c = '@';
      }
      else if (map_fov_brute[x - center_x + radius][y - center_y + radius])
      {
        if (map[x][y] != 0)
          c = '#';
        else
          c = '.';
      }
      else
      {
        if (map[x][y] != 0)
          c = 'X';
        else
          c = 'x';
      }
      putchar(c);
    }

    putchar(' ');

    for (x = center_x - radius; x <= center_x + radius; x++)
    {
      if ((x == center_x) && (y == center_y))
      {
        c = '@';
      }
      else if (map_fov_beam[x - center_x + radius][y - center_y + radius])
      {
        if (map[x][y] != 0)
          c = '#';
        else
          c = '.';
      }
      else
      {
        if (map[x][y] != 0)
          c = 'X';
        else
          c = 'x';
      }
      putchar(c);
    }

    putchar(' ');

    for (x = center_x - radius; x <= center_x + radius; x++)
    {
      if ((x == center_x) && (y == center_y))
      {
        c = '@';
      }
      else if (map_fov_recursive[x - center_x + radius][y - center_y + radius])
      {
        if (map[x][y] != 0)
          c = '#';
        else
          c = '.';
      }
      else
      {
        if (map[x][y] != 0)
          c = 'X';
        else
          c = 'x';
      }
      putchar(c);
    }

    putchar('\n');
  }

  array2_delete(map, 2 * radius + 1, 2 * radius + 1);
  map = NULL;
  array2_delete(map_fov_brute, 2 * radius + 1, 2 * radius + 1);
  map_fov_brute = NULL;
  array2_delete(map_fov_beam, 2 * radius + 1, 2 * radius + 1);
  map_fov_beam = NULL;
  array2_delete(map_fov_recursive, 2 * radius + 1, 2 * radius + 1);
  map_fov_recursive = NULL;

  return 0;
}

int
main(int argc, char *argv[])
{
  srand((unsigned int) time(NULL));

  return one_random_map();
}
