#include <stdio.h>
/* rand, abs */
#include <stdlib.h>
/* time */
#include <time.h>

#include "array.h"
#include "digital-fov-other.h"
#include "digital-fov.h"
#include "util-example.h"

static int dump_map(int **map, int map_size_x, int map_size_y,
                    int ax, int ay, int bx, int by);
static int random_map_this_size(int **map,
                                int map_size_x_max, int map_size_y_max,
                                int map_size_x, int map_size_y,
                                int wall_rarity);
static int verify_match_one_fov(int **map, int map_size_x, int map_size_y,
                                int **map_fov_brute,
                                int **map_fov_beam,
                                int **map_fov_recursive,
                                int center_x, int center_y, int radius);
static int verify_match(int **map, int map_size_x_max, int map_size_y_max,
                        int **map_fov_brute,
                        int **map_fov_beam,
                        int **map_fov_recursive,
                        int radius_max);
static int verify_empty_map(int **map, int map_size_x_max, int map_size_y_max,
                            int **map_fov_brute,
                            int **map_fov_beam,
                            int **map_fov_recursive,
                            int radius_max);
static int verify_full_map(int **map, int map_size_x_max, int map_size_y_max,
                           int **map_fov_brute,
                           int **map_fov_beam,
                           int **map_fov_recursive,
                           int radius_max);
static int verify_radius(int **map, int map_size_x_max, int map_size_y_max,
                         int **map_fov0,
                         int **map_fov1,
                         int radius_max);
static int verify_symmetry(int **map, int map_size_x_max, int map_size_y_max);
static int verify(int **map, int map_size_x_max, int map_size_y_max,
                  int ***map_fovs,
                  int radius_max);

static int
dump_map(int **map, int map_size_x, int map_size_y,
         int ax, int ay, int bx, int by)
{
  int x;
  int y;
  int c;

  if (map == NULL)
    return 1;
  if (map_size_x <= 0)
    return 1;
  if (map_size_y <= 0)
    return 1;

  if (!grid_is_illegal(ax, ay, map_size_x, map_size_y))
  {
    printf("; %s: (%d, %d)\n",
           (map[ax][ay] != 0) ? "A": "a",
           ax, ay);
  }
  if (!grid_is_illegal(bx, by, map_size_x, map_size_y))
  {
    printf("; %s: (%d, %d)\n",
           (map[bx][by] != 0) ? "B": "b",
           bx, by);
  }

  for (y = 0; y < map_size_y; y++)
  {
    for (x = 0; x < map_size_x; x++)
    {
      if ((x == ax) && (y == ay))
      {
        if (map[x][y] != 0)
          c = 'A';
        else
          c = 'a';
      }
      else if ((x == bx) && (y == by))
      {
        if (map[x][y] != 0)
          c = 'B';
        else
          c = 'b';
      }
      else
      {
        if (map[x][y] != 0)
          c = '#';
        else
          c = '.';
      }

      putchar(c);
    }
    putchar('\n');
  }

  return 0;
}

static int
random_map_this_size(int **map,
                     int map_size_x_max, int map_size_y_max,
                     int map_size_x, int map_size_y,
                     int wall_rarity)
{
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if ((map_size_x <= 0) || (map_size_x > map_size_x_max))
    return 1;
  if ((map_size_y <= 0) || (map_size_y > map_size_y_max))
    return 1;

  for (x = 0; x < map_size_x; x++)
  {
    for (y = 0; y < map_size_y; y++)
    {
      if ((wall_rarity > 0) && (rand() % wall_rarity == 0))
        map[x][y] = 1;
      else
        map[x][y] = 0;
    }
  }

  return 0;
}

static int
verify_match_one_fov(int **map, int map_size_x, int map_size_y,
                     int **map_fov_brute,
                     int **map_fov_beam,
                     int **map_fov_recursive,
                     int center_x, int center_y, int radius)
{
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_fov_brute == NULL)
    return 1;
  if (map_fov_beam == NULL)
    return 1;
  if (map_fov_recursive == NULL)
    return 1;
  if (radius < 0)
    return 1;

  digital_fov_brute(map, map_size_x, map_size_y,
                    map_fov_brute, center_x, center_y, radius);
  digital_fov_beam(map, map_size_x, map_size_y,
                   map_fov_beam, center_x, center_y, radius);
  digital_fov(map, map_size_x, map_size_y,
              map_fov_recursive, center_x, center_y, radius);

  for (x = center_x - radius; x <= center_x + radius; x++)
  {
    for (y = center_y - radius; y <= center_y + radius; y++)
    {
      if (map_fov_beam[x - center_x + radius][y - center_y + radius]
          != map_fov_brute[x - center_x + radius][y - center_y + radius])
      {
        printf("; beam and brute mismatch\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, beam = %d, brute = %d\n",
               radius,
               map_fov_beam[x - center_x + radius][y - center_y + radius],
               map_fov_brute[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (map_fov_recursive[x - center_x + radius][y - center_y + radius]
          != map_fov_brute[x - center_x + radius][y - center_y + radius])
      {
        printf("; recursive and brute mismatch\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, recursive = %d, brute = %d\n",
               radius,
               map_fov_recursive[x - center_x + radius][y - center_y + radius],
               map_fov_brute[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (digital_los(map, map_size_x, map_size_y,
                      center_x, center_y, x, y)
          != map_fov_brute[x - center_x + radius][y - center_y + radius])
      {
        printf("; los and brute mismatch\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, los = %d, brute = %d\n",
               radius,
               digital_los(map, map_size_x, map_size_y,
                           center_x, center_y, x, y),
               map_fov_brute[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }
    }
  }

  return 0;
}

static int
verify_match(int **map, int map_size_x_max, int map_size_y_max,
             int **map_fov_brute,
             int **map_fov_beam,
             int **map_fov_recursive,
             int radius_max)
{
  int center_x;
  int center_y;
  int radius;
  int map_size_x;
  int map_size_y;
  int wall_rarity;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov_brute == NULL)
    return 1;
  if (map_fov_beam == NULL)
    return 1;
  if (map_fov_recursive == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  printf("; check implementations match\n");

  map_size_x = 61;
  map_size_y = 61;
  for (wall_rarity = 3; wall_rarity <= 9; wall_rarity += 3)
  {
    printf(";; wall_rarity = %d\n", wall_rarity);

    if (random_map_this_size(map, map_size_x_max, map_size_y_max,
                             map_size_x, map_size_y, wall_rarity) != 0)
      return 1;

    for (radius = 0; radius <= 30; radius++)
    {
      if (radius > radius_max)
        break;

      printf(";;; radius = %d\n", radius);

      for (center_x = 0; center_x < map_size_x; center_x++)
      {
        for (center_y = 0; center_y < map_size_y; center_y++)
        {
          if (verify_match_one_fov(map, map_size_x, map_size_y,
                                   map_fov_brute,
                                   map_fov_beam,
                                   map_fov_recursive,
                                   center_x, center_y, radius) != 0)
            return 1;
        }
      }
    }
  }

  return 0;
}

static int
verify_empty_map(int **map, int map_size_x_max, int map_size_y_max,
                 int **map_fov_brute,
                 int **map_fov_beam,
                 int **map_fov_recursive,
                 int radius_max)
{
  int center_x;
  int center_y;
  int radius;
  int map_size_x;
  int map_size_y;
  int x;
  int y;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov_brute == NULL)
    return 1;
  if (map_fov_beam == NULL)
    return 1;
  if (map_fov_recursive == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  printf("; check empty map\n");

  map_size_x = 61;
  map_size_y = 61;
  center_x = map_size_x / 2;
  center_y = map_size_y / 2;
  radius = 30;

  if (radius > radius_max)
    return 1;
  if (random_map_this_size(map, map_size_x_max, map_size_y_max,
                           map_size_x, map_size_y, 0) != 0)
    return 1;

  digital_fov_brute(map, map_size_x, map_size_y,
                    map_fov_brute, center_x, center_y, radius);
  digital_fov_beam(map, map_size_x, map_size_y,
                   map_fov_beam, center_x, center_y, radius);
  digital_fov(map, map_size_x, map_size_y,
              map_fov_recursive, center_x, center_y, radius);

  for (x = center_x - radius; x <= center_x + radius; x++)
  {
    for (y = center_y - radius; y <= center_y + radius; y++)
    {
      if (map_fov_brute[x - center_x + radius][y - center_y + radius]
          == 0)
      {
        printf("; brute is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, brute = %d\n",
               radius,
               map_fov_brute[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (map_fov_beam[x - center_x + radius][y - center_y + radius]
          == 0)
      {
        printf("; beam is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, beam = %d\n",
               radius,
               map_fov_beam[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (map_fov_recursive[x - center_x + radius][y - center_y + radius]
          == 0)
      {
        printf("; recursive is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, beam = %d\n",
               radius,
               map_fov_recursive[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (digital_los(map, map_size_x, map_size_y,
                      center_x, center_y, x, y)
          == 0)
      {
        printf("; los is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, los = %d\n",
               radius,
               digital_los(map, map_size_x, map_size_y,
                           center_x, center_y, x, y));
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }
    }
  }

  return 0;
}

static int
verify_full_map(int **map, int map_size_x_max, int map_size_y_max,
                int **map_fov_brute,
                int **map_fov_beam,
                int **map_fov_recursive,
                int radius_max)
{
  int center_x;
  int center_y;
  int radius;
  int map_size_x;
  int map_size_y;
  int x;
  int y;
  int c;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov_brute == NULL)
    return 1;
  if (map_fov_beam == NULL)
    return 1;
  if (map_fov_recursive == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  printf("; check full map\n");

  map_size_x = 61;
  map_size_y = 61;
  center_x = map_size_x / 2;
  center_y = map_size_y / 2;
  radius = 30;

  if (radius > radius_max)
    return 1;
  if (random_map_this_size(map, map_size_x_max, map_size_y_max,
                           map_size_x, map_size_y, 1) != 0)
    return 1;

  digital_fov_brute(map, map_size_x, map_size_y,
                    map_fov_brute, center_x, center_y, radius);
  digital_fov_beam(map, map_size_x, map_size_y,
                   map_fov_beam, center_x, center_y, radius);
  digital_fov(map, map_size_x, map_size_y,
              map_fov_recursive, center_x, center_y, radius);

  for (x = center_x - radius; x <= center_x + radius; x++)
  {
    for (y = center_y - radius; y <= center_y + radius; y++)
    {
      if (distance(center_x, center_y, x, y) <= 1)
        c = 1;
      else
        c = 0;

      if (map_fov_brute[x - center_x + radius][y - center_y + radius]
          != c)
      {
        printf("; brute is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, brute = %d\n",
               radius,
               map_fov_brute[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (map_fov_beam[x - center_x + radius][y - center_y + radius]
          != c)
      {
        printf("; beam is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, beam = %d\n",
               radius,
               map_fov_beam[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (map_fov_recursive[x - center_x + radius][y - center_y + radius]
          != c)
      {
        printf("; recursive is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, beam = %d\n",
               radius,
               map_fov_recursive[x - center_x + radius][y - center_y + radius]);
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }

      if (digital_los(map, map_size_x, map_size_y,
                      center_x, center_y, x, y)
          != c)
      {
        printf("; los is wrong\n");
        printf("; center: (%d, %d)\n", center_x, center_y);
        printf("; target: (%d, %d)\n", x, y);
        printf("; radius = %d, los = %d\n",
               radius,
               digital_los(map, map_size_x, map_size_y,
                           center_x, center_y, x, y));
        dump_map(map, map_size_x, map_size_y,
                 center_x, center_y, x, y);
        return 1;
      }
    }
  }

  return 0;
}

static int
verify_radius(int **map, int map_size_x_max, int map_size_y_max,
              int **map_fov0,
              int **map_fov1,
              int radius_max)
{
  int center_x;
  int center_y;
  int radius;
  int map_size_x;
  int map_size_y;
  int wall_rarity;
  int x;
  int y;
  int **map_fov_old = NULL;
  int **map_fov_new = NULL;
  int **map_fov_temp = NULL;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov0 == NULL)
    return 1;
  if (map_fov1 == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  printf("; check radius independence\n");

  map_size_x = map_size_x_max;
  map_size_y = map_size_y_max;
  center_x = map_size_x / 2;
  center_y = map_size_y / 2;
  map_fov_old = map_fov0;
  map_fov_new = map_fov1;

  for (wall_rarity = 2; wall_rarity <= 10; wall_rarity++)
  {
    if (random_map_this_size(map, map_size_x_max, map_size_y_max,
                             map_size_x, map_size_y, wall_rarity) != 0)
      return 1;

    digital_fov(map, map_size_x, map_size_y,
                map_fov_old, center_x, center_y, 0);

    for (radius = 0; radius <= radius_max - 1; radius++)
    {
      if (radius + 1 > radius_max)
        break;

      digital_fov(map, map_size_x, map_size_y,
                  map_fov_new, center_x, center_y, radius + 1);

      for (x = center_x - radius; x <= center_x + radius; x++)
      {
        for (y = center_y - radius; y <= center_y + radius; y++)
        {
          if (map_fov_new[x - center_x + (radius + 1)][y - center_y + (radius + 1)]
              != map_fov_old[x - center_x + radius][y - center_y + radius])
          {
            printf("; FOV depends on radius\n");
            printf("; center: (%d, %d)\n", center_x, center_y);
            printf("; target: (%d, %d)\n", x, y);
            printf("; old radius = %d, new radius = %d\n",
                   radius, radius + 1);
            printf("; map_fov_old = %d, map_fov_new = %d\n",
                   map_fov_old[x - center_x + radius][y - center_y + radius],
                   map_fov_new[x - center_x + (radius + 1)][y - center_y + (radius + 1)]);
            dump_map(map, map_size_x, map_size_y,
                     center_x, center_y, x, y);
            return 1;
          }
        }
      }

      map_fov_temp = map_fov_old;
      map_fov_old = map_fov_new;
      map_fov_new = map_fov_temp;
    }
  }

  return 0;
}

static int
verify_symmetry(int **map, int map_size_x_max, int map_size_y_max)
{
  int map_size_x;
  int map_size_y;
  int wall_rarity;
  int ax;
  int ay;
  int bx;
  int by;

  if (map == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;

  printf("; check symmetry\n");

  map_size_x = 61;
  map_size_y = 61;
  for (wall_rarity = 3; wall_rarity <= 9; wall_rarity += 3)
  {
    printf(";; wall_rarity = %d\n", wall_rarity);

    if (random_map_this_size(map, map_size_x_max, map_size_y_max,
                             map_size_x, map_size_y, wall_rarity) != 0)
      return 1;

    for (ax = 0; ax < map_size_x; ax++)
    {
      for (ay = 0; ay < map_size_y; ay++)
      {
        for (bx = ax; bx < map_size_x; bx++)
        {
          for (by = ay; by < map_size_y; by++)
          {
            if (digital_los(map, map_size_x, map_size_y,
                            ax, ay, bx, by)
                != digital_los(map, map_size_x, map_size_y,
                               bx, by, ax, ay))
            {
              printf("; los is not symmetric\n");
              printf("; (ax, ay): (%d, %d)\n", ax, ay);
              printf("; (bx, by): (%d, %d)\n", bx, by);
              printf("; los(ax, ay, bx, by) = %d\n",
                     digital_los(map, map_size_x, map_size_y,
                                 ax, ay, bx, by));
              printf("; los(bx, by, ax, ay) = %d\n",
                     digital_los(map, map_size_x, map_size_y,
                                 bx, by, ax, ay));
              dump_map(map, map_size_x, map_size_y,
                       ax, ay, bx, by);
              return 1;
            }
          }
        }
      }
    }
  }
 
  return 0;
}

static int
verify(int **map, int map_size_x_max, int map_size_y_max,
       int ***map_fovs,
       int radius_max)
{
  if (verify_match(map, map_size_x_max, map_size_y_max,
                   map_fovs[0],
                   map_fovs[1],
                   map_fovs[2],
                   radius_max) != 0)
    return 1;
  if (verify_empty_map(map, map_size_x_max, map_size_y_max,
                       map_fovs[0],
                       map_fovs[1],
                       map_fovs[2],
                       radius_max) != 0)
    return 1;
  if (verify_full_map(map, map_size_x_max, map_size_y_max,
                      map_fovs[0],
                      map_fovs[1],
                      map_fovs[2],
                      radius_max) != 0)
    return 1;
  if (verify_radius(map, map_size_x_max, map_size_y_max,
                    map_fovs[0],
                    map_fovs[1],
                    radius_max) != 0)
    return 1;
  if (verify_symmetry(map, map_size_x_max, map_size_y_max) != 0)
    return 1;

  return 0;
}

int
main(int argc, char *argv[])
{
  int map_size_x_max;
  int map_size_y_max;
  int radius_max;
  int error_found;
  int **map = NULL;
  int ***map_fovs = NULL;

  radius_max = 300;
  map_size_x_max = 2 * radius_max + 1;
  map_size_y_max = 2 * radius_max + 1;

  map = array2_new(map_size_x_max, map_size_y_max);
  if (map == NULL)
  {
    return 1;
  }
  map_fovs = array3_new(3, 2 * radius_max + 1, 2 * radius_max + 1);
  if (map_fovs == NULL)
  {
    array2_delete(map, map_size_x_max, map_size_y_max);
    map = NULL;
    return 1;
  }

  srand((unsigned int) time(NULL));

  error_found = 0;
  if (verify(map, map_size_x_max, map_size_y_max,
             map_fovs,
             radius_max) != 0)
    error_found = 1;

  array2_delete(map, map_size_x_max, map_size_y_max);
  map = NULL;
  array3_delete(map_fovs, 3, 2 * radius_max + 1, 2 * radius_max + 1);
  map_fovs = NULL;

  return error_found;
}
