#include <stdio.h>
/* rand, abs */
#include <stdlib.h>
/* time */
#include <time.h>
/* strcmp */
#include <string.h>
#include <curses.h>

#include "array.h"
#include "digital-fov-other.h"
#include "digital-fov.h"
#include "util-example.h"

#define RANDOM_MAP_RADIUS 30
#define RANDOM_MAP_WALL_RARITY 6

static void curses_cleanup(void);
static int walk_around_draw(int **map, int map_size_x, int map_size_y,
                            int center_x, int center_y,
                            int **map_char, WINDOW *scr,
                            int **map_fov, int radius, int draw_non_visible,
                            const char *fov_func_name);
static int walk_around_main_loop(int **map, int map_size_x, int map_size_y,
                                 int center_x, int center_y,
                                 int **map_char, WINDOW *scr,
                                 const char *file_name,
                                 int map_size_x_max, int map_size_y_max,
                                 int **map_fov, int radius_max);
static int walk_around(int **map, int map_size_x_max, int map_size_y_max,
                       int **map_char, const char *file_name,
                       int **map_fov, int radius_max);

static
void curses_cleanup(void)
{
  clear();
  refresh();
  nocbreak();
  echo();
  endwin();
}

/* the caller must calculate FOV before calling */
static int
walk_around_draw(int **map, int map_size_x, int map_size_y,
                 int center_x, int center_y,
                 int **map_char, WINDOW *scr,
                 int **map_fov, int radius, int draw_non_visible,
                 const char *fov_func_name)
{
  int x;
  int y;
  int i;
  int j;
  int c;
  int scr_size_x;
  int scr_size_y;
  int attr;
  char buf[128];

  if (map == NULL)
    return 1;
  if (map_char == NULL)
    return 1;
  if (map_size_x <= 0)
    return 1;
  if (map_size_y <= 0)
    return 1;
  if (scr == NULL)
    return 1;
  if (map_fov == NULL)
    return 1;
  if (radius < 0)
    return 1;
  if (fov_func_name == NULL)
    return 1;

  getmaxyx(scr, scr_size_y, scr_size_x);

  erase();

  for (j = 0; j < scr_size_y - 1; j++)
  {
    for (i = 0; i < scr_size_x; i++)
    {
      x = i - scr_size_x / 2 + center_x;
      y = j - scr_size_y / 2 + center_y;

      if (grid_is_illegal(x, y, map_size_x, map_size_y))
        continue;

      c = map_char[x][y];
      attr = COLOR_PAIR(7);
      if ((x == center_x) && (y == center_y))
      {
        c = '@';
        attr = COLOR_PAIR(1) | A_BOLD;
      }
      else if ((distance(center_x, center_y, x, y) <= radius)
               && (map_fov[x - center_x + radius][y - center_y + radius]))
      {
        if (((x - center_x) == (y - center_y))
            || ((x - center_x) == -(y - center_y)))
          attr = COLOR_PAIR(5);
        else if (map[x][y] != 0)
          attr = COLOR_PAIR(3);
        else
          attr = COLOR_PAIR(7);
      }
      else
      {
        if (!draw_non_visible)
          continue;
        if (map[x][y] != 0)
          attr = COLOR_PAIR(4);
        else
          attr = COLOR_PAIR(8) | A_BOLD;
      }

      attron(attr);
      mvaddch(j, i, c);
      attroff(attr);
    }
  }

  sprintf(buf, "@(%d, %d)  radius= %d  %s",
          center_x, center_y,
          radius,
          fov_func_name);
  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(scr_size_y - 1, 1, buf);
  attroff(attr);

  move(scr_size_y - 1, 0);
  refresh();

  return 0;
}

static int
walk_around_main_loop(int **map, int map_size_x, int map_size_y,
                      int center_x, int center_y,
                      int **map_char, WINDOW *scr,
                      const char *file_name,
                      int map_size_x_max, int map_size_y_max,
                      int **map_fov, int radius_max)
{
  int c;
  int should_quit;
  int fov_func_type;
  int radius;
  int draw_non_visible;
  int circle_fov;
  int (*fov_func)(int **, int, int,
                  int **,
                  int, int, int) = NULL;
  const char *fov_func_name = NULL;

  if (map == NULL)
    return 1;
  if (map_char == NULL)
    return 1;
  if (map_size_x <= 0)
    return 1;
  if (map_size_y <= 0)
    return 1;
  if (scr == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  fov_func_type = 0;
  radius = 10;
  draw_non_visible = 1;
  circle_fov = 0;

  should_quit = 0;
  while (!should_quit)
  {
    if (map_size_x_max <= 0)
      break;
    if (map_size_y_max <= 0)
      break;
    if (center_x < 0)
      center_x = 0;
    if (center_x > map_size_x - 1)
      center_x = map_size_x - 1;
    if (center_y < 0)
      center_y = 0;
    if (center_y > map_size_y - 1)
      center_y = map_size_y - 1;
    if (radius < 0)
      radius = 0;
    if (radius > radius_max)
      radius = radius_max;
    if (fov_func_type < 0)
      fov_func_type = 0;
    if (fov_func_type > 3)
      fov_func_type = 3;

    switch (fov_func_type)
    {
    case 0:
      fov_func = &digital_fov;
      fov_func_name = "recursive";
      break;
    case 1:
      fov_func = &digital_fov_brute;
      fov_func_name = "brute";
      break;
    case 2:
      fov_func = &digital_fov_beam;
      fov_func_name = "beam";
      break;
    case 3:
      fov_func = &digital_fov_los;
      fov_func_name = "los";
      break;
    default:
      /* something better than NULL */
      fov_func = &dummy_fov;
      fov_func_name = "dummy";
      break;
    }

    (*fov_func)(map,  map_size_x, map_size_y,
                map_fov,
                center_x, center_y, radius);
    if (circle_fov)
      trim_circle(map_fov, radius);

    walk_around_draw(map, map_size_x, map_size_y,
                     center_x, center_y,
                     map_char, scr,
                     map_fov, radius, draw_non_visible,
                     fov_func_name);

    c = getch();
    switch (c)
    {
    case 'q':
      should_quit = 1;
      break;
    case '7':
    case 'y':
      center_x += -1;
      center_y += -1;
      break;
    case '8':
    case 'k':
    case KEY_UP:
      center_x += 0;
      center_y += -1;
      break;
    case '9':
    case 'u':
      center_x += 1;
      center_y += -1;
      break;
    case '4':
    case 'h':
    case KEY_LEFT:
      center_x += -1;
      center_y += 0;
      break;
    case '6':
    case 'l':
    case KEY_RIGHT:
      center_x += 1;
      center_y += 0;
      break;
    case '1':
    case 'b':
      center_x += -1;
      center_y += 1;
      break;
    case '2':
    case 'j':
    case KEY_DOWN:
      center_x += 0;
      center_y += 1;
      break;
    case '3':
    case 'n':
      center_x += 1;
      center_y += 1;
      break;
    case 'r':
      if (file_name != NULL)
      { 
        read_map_from_file(file_name,
                           map, map_size_x_max, map_size_y_max,
                           &map_size_x, &map_size_y,
                           NULL, NULL, map_char,
                           1);
        if (map_size_x <= 0)
        {
          should_quit = 1;
        break;
        }
        if (map_size_y <= 0)
        {
          should_quit = 1;
        break;
        }
      }
      else
      {
        random_map(map, map_size_x_max, map_size_y_max,
                   &map_size_x, &map_size_y,
                   NULL, NULL, map_char,
                   RANDOM_MAP_RADIUS, RANDOM_MAP_WALL_RARITY);
      }
      break;
    case 'm':
      if (draw_non_visible)
        draw_non_visible = 0;
      else
        draw_non_visible = 1;
      break;
    case '+':
      radius++;
      break;
    case '-':
      radius--;
      break;
    case 'c':
      if (circle_fov)
        circle_fov = 0;
      else
        circle_fov = 1;
      break;
    case 'i':
      fov_func_type = (fov_func_type + 1) % 4;
      break;
    default:
      break;
    }
  }
  
  return 0;
}

static int
walk_around(int **map, int map_size_x_max, int map_size_y_max,
            int **map_char, const char *file_name,
            int **map_fov, int radius_max)
{
  int map_size_x;
  int map_size_y;
  int center_x;
  int center_y;
  WINDOW *scr = NULL;

  if (map == NULL)
    return 1;
  if (map_char == NULL)
    return 1;
  if (map_size_x_max <= 0)
    return 1;
  if (map_size_y_max <= 0)
    return 1;
  if (map_fov == NULL)
    return 1;
  if (radius_max < 0)
    return 1;

  if (file_name != NULL)
  {  
    if (read_map_from_file(file_name,
                           map, map_size_x_max, map_size_y_max,
                           &map_size_x, &map_size_y,
                           &center_x, &center_y, map_char,
                           0) != 0)
    {
      fprintf(stderr, "walk_around: read_map_from_file failed\n");
      return 1;
    }
    if ((map_size_x <= 0) || (map_size_y <= 0))
    {
      fprintf(stderr, "walk_around: strange map size %d x %d\n",
              map_size_x, map_size_y);
      return 1;
    }
  }
  else
  {
    random_map(map, map_size_x_max, map_size_y_max,
               &map_size_x, &map_size_y,
               &center_x, &center_y, map_char,
               RANDOM_MAP_RADIUS, RANDOM_MAP_WALL_RARITY);
  }
  
  scr = initscr();
  if (scr == NULL)
  {
    fprintf(stderr, "walk_around: initscr failed\n");
    return 1;
  }
  if (atexit(&curses_cleanup) != 0)
  {
    fprintf(stderr, "walk_around: atexit failed\n");
    curses_cleanup();
    return 1;
  }

  start_color();
  noecho();
  cbreak();
  keypad(scr, TRUE);

  init_pair(1, COLOR_RED, COLOR_BLACK);
  init_pair(2, COLOR_GREEN, COLOR_BLACK);
  init_pair(3, COLOR_YELLOW, COLOR_BLACK);
  init_pair(4, COLOR_BLUE, COLOR_BLACK);
  init_pair(5, COLOR_MAGENTA, COLOR_BLACK);
  init_pair(6, COLOR_CYAN, COLOR_BLACK);
  init_pair(7, COLOR_WHITE, COLOR_BLACK);
  init_pair(8, COLOR_BLACK, COLOR_BLACK);

  walk_around_main_loop(map, map_size_x, map_size_y,
                        center_x, center_y,
                        map_char, scr,
                        file_name,
                        map_size_x_max, map_size_y_max,
                        map_fov, radius_max);

  return 0;
}

int
main(int argc, char *argv[])
{
  int i;
  int map_size_x_max;
  int map_size_y_max;
  int radius_max;
  int **map = NULL;
  int **map_char = NULL;
  int **map_fov = NULL;
  char *file_name = NULL;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--map-file"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for %s\n", argv[i]);
        return 1;
      }
      file_name = argv[i + 1];
      i++;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      return 1;
    }
  }

  map_size_x_max = 1001;
  map_size_y_max = 1001;
  radius_max = 50;

  map = array2_new(map_size_x_max, map_size_y_max);
  if (map == NULL)
  {
    return 1;
  }
  map_char = array2_new(map_size_x_max, map_size_y_max);
  if (map_char == NULL)
  {
    array2_delete(map, map_size_x_max, map_size_y_max);
    map = NULL;
    return 1;
  }
  map_fov = array2_new(2 * radius_max + 1, 2 * radius_max + 1);
  if (map_fov == NULL)
  {
    array2_delete(map, map_size_x_max, map_size_y_max);
    map = NULL;
    array2_delete(map_char, map_size_x_max, map_size_y_max);
    map_char = NULL;
    return 1;
  }

  srand((unsigned int) time(NULL));

  walk_around(map, map_size_x_max, map_size_y_max,
              map_char, file_name,
              map_fov, radius_max);

  array2_delete(map, map_size_x_max, map_size_y_max);
  map = NULL;
  array2_delete(map_char, map_size_x_max, map_size_y_max);
  map_char = NULL;
  array2_delete(map_fov, 2 * radius_max + 1, 2 * radius_max + 1);
  map_fov = NULL;

  return 0;
}
